// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <boost/property_tree/json_parser.hpp>
#include <boost/property_tree/ptree.hpp>
#include <chrono>
#include <fstream>
#include <iostream>
#include <pthread.h>
#include <string>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <termios.h>
#include <vector>
#include <xrm.h>
#include <filesystem>

// #define DEBUG
static constexpr uint32_t MAX_REQ_JOB_SLOTS = 9999;
static constexpr uint32_t PLUGIN_COUNT      = 4;
static constexpr uint32_t XRM_PLUGIN_FUNCID = 0;
// Max devices supported per usecase
static constexpr uint32_t MAX_SUPPORTED_DEVICES = 16;
// Relook into need of CMD_SIZE, LAUNCHER_CMD_SIZE
static constexpr uint32_t CMD_SIZE          = 512;
static constexpr uint32_t LAUNCHER_CMD_SIZE = 4096;

static constexpr int32_t XRM_PRECISION_1000000_BIT_MASK(int32_t load) {
  return (load << 8);
}

static constexpr const char* VAR_TMP_AMD   = "/var/tmp/amd/";
static constexpr const char* LOG_FILE_PATH = "/var/tmp/amd/ma35/";
namespace pt                               = boost::property_tree;

// Added to handle CTRL+C while launcher application is active
static std::vector<pid_t> gl_pids;
static pthread_mutex_t    gl_mutex_lock = PTHREAD_MUTEX_INITIALIZER;

// To map virtaul device ids provided in run cmds to physical device ids returned by XRM
typedef struct xrm_dev_list {
  std::vector<int32_t> hw_dev_ids;
  std::vector<int32_t> vt_dev_ids;
  std::vector<int32_t> dev_start_cuidx;
} xrm_dev_list;

// Minimal data needed to execute a single job independently
struct thread_data {
  std::string                         source_filename;
  std::string                         pre_src_cmdline;
  xrm_dev_list*                       pop_dev_list;
  int                                 log_level;
  uint32_t                            job_type;
  uint32_t                            reservation_id;
  void*                               opaque; //Data to be used for callback
  std::function<int(void*, int, int)> callback;
};

struct cuInfo {
  std::vector<uint32_t>    cu_load;
  std::vector<uint32_t>    num_channels;
  std::string              plugin_name;
  std::vector<std::string> kernel_name;
  std::vector<std::string> kernel_alias;
};

enum reserveIdStatus {
  RESERVE_ID_STATUS_UNUSED = -1,
  RESERVE_ID_STATUS_FREE   = 0,
  RESERVE_ID_STATUS_INUSE  = 1,
};

struct jobSlotInfo {
  bool                                 is_ull = false;
  uint32_t                             num_dev_per_cmd;
  uint32_t                             req_job_slots = MAX_REQ_JOB_SLOTS;
  int32_t                              ref_idx       = -1;
  float                                load_factor   = 1.0; // to limit the maximum allowed jobslots per device
  std::string                          json_file_name;
  std::string                          run_commmand;
  std::vector<uint32_t>                start_cu_idx;
  std::vector<std::string>             input_files;
  std::unique_ptr<xrmCuPoolPropertyV2> xrm_cu_pool_prop;

  uint32_t                                  finished_jobs = 0;
  uint32_t                                  allot_job_slots;
  std::vector<int32_t>                      reservation_ids;
  std::vector<reserveIdStatus>              reservation_id_status;
  std::vector<std::unique_ptr<pthread_t>>   thread_ptr;
  std::vector<std::unique_ptr<thread_data>> thread_data_ptr;
  std::vector<xrm_dev_list>                 pop_dev_list_vec;
  pthread_mutex_t                           mutex_lock;
  std::vector<std::string>                  failed_jobs;
};

struct jobSlotAppInfo {
  int32_t                             log_level           = 0; // similar to U30
  int32_t                             last_res_id         = 0;
  bool                                is_dry_run          = false;
  int32_t                             total_jobs_reserved = 0;
  xrmContext*                         xrm_ctx             = nullptr;
  std::unique_ptr<xrmPluginFuncParam> plugin_param_ptr;
  std::vector<jobSlotInfo>            job_info;
  cuInfo cu_info[PLUGIN_COUNT] = {{{}, {}, "xrmMA35MLPlugin", {"ml"}, {"ML_AMA"}}, {{}, {}, "xrmMA35ScalePlugin", {"scaler"}, {"SCALER_AMA"}},
      {{}, {}, "xrmMA35DecodePlugin", {"decoder"}, {"DECODER_AMA"}},
      {{}, {}, "xrmMA35EncodePlugin", {"encoder", "encoder", "lookahead"}, {"ENCODER_TYPE1_AMA", "ENCODER_TYPE2_AMA", "LOOKAHEAD_AMA"}}};

  ~jobSlotAppInfo() {
    // relinquish reserve Ids and free free correspsonding CU reources
    for (size_t files = 0; files < job_info.size(); ++files) {
      pthread_mutex_destroy(&job_info[files].mutex_lock);
      for (size_t i = 0; i < job_info[files].reservation_ids.size(); ++i) {
        xrmCuPoolRelinquish(xrm_ctx, job_info[files].reservation_ids[i]);
      }
    }
    // Destroy XRM context
    xrmDestroyContext(xrm_ctx);
  };
};

// print -h and --help options when called
static void print_help(const char* exe) {
  fprintf(stdout, "\nUsage:-\n");
  fprintf(stdout, "       %s sources_filename_1 cmds_filename_1 [sources_filename_2 cmds_filename_2] ... [sources_filename_n cmds_filename_n]\n", exe);
  fprintf(stdout, "       \t\tOR\n");
  fprintf(stdout, "       %s sources_filename_1 cmds_filename_1 [sources_filename_2 cmds_filename_2] ... [sources_filename_n cmds_filename_n] --log_level [0-2]\n", exe);
  return;
}

// check access for log file creation
static int check_folder_access() {
  struct stat dstat;
  if (stat(LOG_FILE_PATH, &dstat) == -1) {
    mode_t  target_mode = S_IRWXU | S_IRWXG | S_IRWXO;
    int32_t ret         = mkdir(VAR_TMP_AMD, target_mode);
    if (ret == EEXIST) {
      fprintf(stdout, "\nFolder exists skiping mkdir\n");
    } else if (ret != 0) {
      fprintf(stderr, "\nError: Couldn't create %s folder to write logs. Err = %s\n", VAR_TMP_AMD, strerror(errno));
      return 1;
    } else {
      // Fixed folder access issue for multi-user
      chmod(VAR_TMP_AMD, target_mode);
    }
    ret = mkdir(LOG_FILE_PATH, target_mode);
    if (ret == EEXIST) {
      fprintf(stdout, "\nFolder exists skiping mkdir\n");
    } else if (ret != 0) {
      fprintf(stderr, "\nError: Couldn't create %s folder to write logs. Err = %s\n", LOG_FILE_PATH, strerror(errno));
      return 1;
    } else {
      // Fixed folder access issue for multi-user
      chmod(LOG_FILE_PATH, target_mode);
    }
  }
  return 0;
}

int get_ffmpeg_cmd_vt_devs(std::string run_command, xrm_dev_list* pop_dev_list) {
  long dev_id = 0;
  bool dev_list[MAX_SUPPORTED_DEVICES];
  memset(dev_list, false, MAX_SUPPORTED_DEVICES * sizeof(bool));
  int                i = 0;
  std::string        token;
  std::istringstream tokenStream(run_command);
  while (std::getline(tokenStream, token, ' ')) {
    // Add for gstreamer after requirement is freezed
    if (token.find("ama=") != std::string::npos) {
      size_t found = token.find("transcoder");
      if (found != std::string::npos) {
        std::string vt_dev_id = "";
        for (size_t k = found + 10; k < token.size(); k++) {
          if (token[k] >= '0' && token[k] <= '9') {
            vt_dev_id = vt_dev_id + token[k];
          } else {
            break;
          }
        }
        if (vt_dev_id != "") {
          // handle stoi error
          dev_id = stoi(vt_dev_id);
        }

        if ((dev_id < MAX_SUPPORTED_DEVICES) && (dev_id > -1)) {
          if (dev_list[dev_id] == false) {
            pop_dev_list->vt_dev_ids.push_back(int(dev_id));
            dev_list[dev_id] = true;
            i++;
          }
        } else {
          fprintf(stderr, "\nError: Invalid device ID %ld suppled to Amd device command line options for command: %s\n", dev_id, run_command.c_str());
          return -1;
        }
      }
    }
  }

  return 0;
}

// Check if line is non empty or not
int not_empty_line(const std::string& line) {
  for (size_t i = 0; i < line.length(); i++) {
    if (!isspace(line[i])) {
      return 1;
    }
  }
  return 0;
}
// Extaract key value from run commnad file
int key_value(const std::string& line, std::string& key, std::string& value) {
  auto pos = line.find_first_of('=');
  if (pos == std::string::npos)
    return 1;
  std::string temp_key   = line.substr(0, pos - 1);
  std::string temp_value = line.substr(pos + 1);
  if (!not_empty_line(temp_key) || !not_empty_line(temp_value) || temp_key.length() == 0 || temp_value.length() == 0) {
    return 1;
  }
  key   = temp_key.substr(0, temp_key.find_last_not_of(" \n\r\t\f\v") + 1);
  value = temp_value.substr(temp_value.find_first_not_of(" \n\r\t\f\v"));

  return 0;
}

int file_exist(std::string file_path) {
  std::ifstream file(file_path);
  int           ret = 0;
  if (file.is_open()) {
    ret = 1;
  }
  file.close();
  return ret;
}

// Check Valid or invalid arguments in argc
static int parse_arguments(uint32_t argc, const char** argv, jobSlotAppInfo* app_info_ptr) {
  // fprintf (stdout, "\nNo of args passed = : %d\n", argc);
  if (argc < 3) {
    fprintf(stderr, "\nError: Not enough arguments provided\n");
    print_help(argv[0]);
    return 1;
  }
  int input_file_flag = 1;
  for (uint32_t i = 1, j = 0; i < argc; ++i) {
    if (argv[i][0] == '-') {
      if ((argv[i][1] == 'h' && argv[i][2] == '\0') || strcmp("--help", argv[i]) == 0) {
        print_help(argv[0]);
        return 1;
      } else if ((argv[i][1] == 'l' && argv[i][2] == '\0') || strcmp("--log_level", argv[i]) == 0) {
        uint32_t log_level = atoi(argv[i + 1]);
        if (log_level > 2) {
          fprintf(stdout, "\nlog_level(%u) not supported and set to max supported value(2)\n", log_level);
          log_level = 2;
        }
        app_info_ptr->log_level = log_level;
        i++; // fast-forward by one arg. Fix this properly
      } else {
        fprintf(stderr, "\nError: Unknown Option %s\nGet info on usage:\n     %s --help \n   or   \n     %s -h \n", argv[i], argv[0], argv[0]);
        return 1;
      }
    } else {
      std::string file_path(argv[i]);
      if (file_path.find(".txt") != std::string::npos) {
        std::ifstream file;
        if (input_file_flag) {
          if (file_exist(argv[i])) {
            app_info_ptr->job_info.push_back(jobSlotInfo());
            if (pthread_mutex_init(&app_info_ptr->job_info[app_info_ptr->job_info.size() - 1].mutex_lock, NULL) != 0) {
              // fprintf (stdout,"\n Mutex init has failed\n");
              return 1;
            }
          } else {
            fprintf(stderr, "\nError: %s file does not exist and continuing with other files, if any\n", argv[i]);
            input_file_flag = 0;
            continue;
          }
        } else {
          if ((!file_exist(argv[i])) || (!file_exist(argv[i - 1]))) {
            if (file_exist(argv[i - 1])) {
              app_info_ptr->job_info.pop_back();
            }
            if (!file_exist(argv[i])) {
              fprintf(stderr, "\nError: %s file does not exist and continuing with other files, if any\n", argv[i]);
            }
            input_file_flag = 1;
            continue;
          }
        }
        std::string line, value;
        try {
          file.open(argv[i]);

          while (!file.eof()) {
            getline(file, line);
            if (line.length()) {
              auto idx = line.find_first_not_of(" \n\r\t\f\v");
              if (line[idx] != '#' && idx != std::string::npos) {
                if (not_empty_line(line)) {
                  if (idx != 0) {
                    line = line.substr(idx);
                  }
                  line.resize(line.find_last_not_of(" \n\r\t\f\v") + 1);
                  if (input_file_flag) {
                    std::string        token;
                    std::istringstream InputStream(line);
                    int32_t            all_input_exist = 1;
                    while (std::getline(InputStream, token, ' ')) {
                      if (!file_exist(token)) {
                        all_input_exist = 0;
                      }
                    }
                    if (all_input_exist) {
                      app_info_ptr->job_info[j].input_files.push_back(line);
                    } else {
                      fprintf(stderr, "\nWarning: In \"%s\" file \"%s\" does not exist and skipping it!!!\n", argv[i], line.c_str());
                    }
                  } else {
                    std::string key;
                    if (key_value(line, key, value)) {
                      fprintf(stderr, "\nWarning: In file %s Some entries are not correct.\n", argv[i]);
                      continue;
                    } else {
                      // fprintf (stdout,  "\nkey %s\nvalue %s\n", key.c_str(), value.c_str());
                      if (key == "job_description") {
                        app_info_ptr->job_info[j].json_file_name = value;
                      } else if (key == "cmdline") {
                        app_info_ptr->job_info[j].run_commmand = value;
                      }
                    }
                  }
                }
              }
            }
          }
          file.close();

        } catch (const std::ifstream::failure& e) {
          std::cerr << "Exception opening/reading/closing file\n";
          return 1;
        }
        // check no of input files >0
        if ((app_info_ptr->job_info[j].input_files.size() == 0 || !file_exist(app_info_ptr->job_info[j].json_file_name) ||
                app_info_ptr->job_info[j].run_commmand == "") &&
            !input_file_flag) {
          fprintf(stderr, "\n============================\n");
          fprintf(stderr, "Error : For Job Type\n");
          fprintf(stderr, "File name: %s\n", argv[i]);
          if (app_info_ptr->job_info[j].input_files.size() == 0) {
            fprintf(stderr, "No input files found\n");
          }
          if (!file_exist(app_info_ptr->job_info[j].json_file_name)) {
            fprintf(stderr, "Json file %s not found\n", app_info_ptr->job_info[j].json_file_name.c_str());
          }
          if (app_info_ptr->job_info[j].run_commmand == "") {
            fprintf(stderr, "No Run command found\n");
          }
          fprintf(stderr, "\n============================\n");
          app_info_ptr->job_info.pop_back();
          input_file_flag = 1;
          continue;
        }

        if (input_file_flag) {
          input_file_flag = 0;
        } else {
          input_file_flag = 1;
          ++j;
        }
      } else {
        fprintf(stderr, "\nError: Unknown file-Fomat %s\nGet info on usage:\n     %s --help \n   or   \n     %s -h \n", argv[i], argv[0], argv[0]);
        return 1;
      }
    }
  }
  if (input_file_flag != 1) {
    fprintf(stderr, "\nError: Missing source.txt/Run.txt, exiting......\n");
    return 1;
  }

  if (app_info_ptr->job_info.size() < 1) {
    fprintf(stderr, "\nError: No Input files provided / Input files not found , exiting......\n");
    return 1;
  }

#ifdef DEBUG
  fprintf(stdout,
      "\n###############################################################"
      "\n\n");
#endif

  return 0;
}

// Get fileName without extension and path
static __inline std::string getFileName(std::string file_path) {
  std::filesystem::path fs_path(file_path);
  return fs_path.stem();
}

// This function needs to be called once for each CU request from XRM, which populates the XRM CU properties for the CU.
static void fill_xrm_cu_props(
    xrmCuPoolPropertyV2* xrm_cu_pool_prop, uint32_t cu_load, uint32_t cu_index, const std::string& kernel_name, const std::string& kernel_alias, uint32_t dev_idx) {
  uint64_t deviceInfoContraintType = XRM_DEVICE_INFO_CONSTRAINT_TYPE_VIRTUAL_DEVICE_INDEX;
  strcpy(xrm_cu_pool_prop->cuListProp.cuProps[cu_index].kernelName, kernel_name.c_str());
  strcpy(xrm_cu_pool_prop->cuListProp.cuProps[cu_index].kernelAlias, kernel_alias.c_str());
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].devExcl    = false;
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].deviceInfo = (dev_idx << XRM_DEVICE_INFO_DEVICE_INDEX_SHIFT) |
      (deviceInfoContraintType << XRM_DEVICE_INFO_CONSTRAINT_TYPE_SHIFT);
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].requestLoad = XRM_PRECISION_1000000_BIT_MASK(cu_load);
  return;
}

// This function poupulates the Logical Grouping Information needed to support Slice Id, Slice Pinning & ULL cases.
// Needed to request VC8000E, XAV1 Encoder and corresponding Lookahead CU allocations from XRM
static void update_lg_info(xrmCuPoolPropertyV2* xrm_cu_pool_prop, uint32_t cu_load, bool is_ull, uint32_t& cu_index, const std::string& kernel_alias,
    uint32_t lookahead_load, int32_t pseudo_logic_group_idx) {
  // As per the XRM 2nd drop design, constraint info should be set for each CU to meet Slice aware, Slice Pinning & ULL requirements
  // For ULL, XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_REAL_LOGIC_GROUP_ID should be set with actual groupId
  // For Slice pinning of Encoder and Lookahead, XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_PSEUDO_LOGIC_GROUP_ID should be set with psuedo groupId (only if Lookahead IP is enabled)
  // Slice Aware, XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE should be used with appropriate groupType (lookahead is disabled)
  uint32_t constraint_info                                                     = is_ull ?
                                                          XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_REAL_LOGIC_GROUP_ID :
                                                          (lookahead_load ? XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_PSEUDO_LOGIC_GROUP_ID : XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE);
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.constraintInfo = constraint_info;
  int32_t real_logic_group_idx                                                 = 0;
  xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupId        = is_ull ? real_logic_group_idx : pseudo_logic_group_idx;
  // Group Type is valid only if constraint_info = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE
  if (xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.constraintInfo == XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE)
    strcpy(xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupType, (kernel_alias == "ENCODER_TYPE1_AMA") ? "ENCODER_TYPE1" : "ENCODER_TYPE2");

  // Request for additional CUs of Encoder & lookahead from Slice 1 if it is a ULL encode channel
  if (is_ull) {
    // make 2 half loads and query from both slices for ULL
    cu_load = (cu_load + 1) >> 1;
    // Overwrite with updated cu_load
    xrm_cu_pool_prop->cuListProp.cuProps[cu_index].requestLoad = XRM_PRECISION_1000000_BIT_MASK(cu_load);
    if (kernel_alias == "ENCODER_TYPE1_AMA") {
      // Default groupID is zero. increment by 1, if it is XAV1 encoder CU.
      xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupId++;
    } else if (kernel_alias == "LOOKAHEAD_AMA") {
      // Lookahead CU group ID should be same as corresponding Encoder CU group ID. So populating from previously populated Encoder CU properties.
      xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupId = xrm_cu_pool_prop->cuListProp.cuProps[cu_index - 2].logicGroupInfo.groupId;
    }
    memcpy(&xrm_cu_pool_prop->cuListProp.cuProps[cu_index + 1], &xrm_cu_pool_prop->cuListProp.cuProps[cu_index], sizeof(xrmCuPropertyV2));
    xrm_cu_pool_prop->cuListProp.cuProps[cu_index + 1].logicGroupInfo.groupId += 2;
    // fprintf (stdout,  "\nGroup id 0: %d\n", xrm_cu_pool_prop->cuListProp.cuProps[cu_index].logicGroupInfo.groupId);
    // fprintf (stdout,  "\nGroup id 1: %d\n", xrm_cu_pool_prop->cuListProp.cuProps[cu_index+1].logicGroupInfo.groupId);
    cu_index += 1;
  }

  return;
}

// Populate XRM cu pool property structs for non zero values of CU loads
static void fill_xrm_props(jobSlotAppInfo* app_info_ptr, int32_t job_index) {
  uint32_t index                                                = 0;
  int32_t  enc_load_idx                                         = 0;
  app_info_ptr->job_info[job_index].xrm_cu_pool_prop->cuListNum = 1;
  int32_t pseudo_logic_group_idx                                = -1;

  // Device loop
  for (uint32_t dev_idx = 0; dev_idx < app_info_ptr->job_info[job_index].num_dev_per_cmd; ++dev_idx) {
    app_info_ptr->job_info[job_index].start_cu_idx.push_back(index);
    // Plugin loop
    for (size_t plugin_idx = 0; plugin_idx < sizeof(app_info_ptr->cu_info) / sizeof(cuInfo); ++plugin_idx) {
      cuInfo* job_cu_info   = app_info_ptr->cu_info;
      bool    is_encoder_cu = (job_cu_info[plugin_idx].kernel_name[0] == "encoder");
      // Channel loop > 1 valid for Encoder CUs only
      for (uint32_t channel_count = 0; channel_count < job_cu_info[plugin_idx].num_channels[dev_idx]; ++channel_count) {
        // load_idx is same as dev_idx, except for encoder CU
        int32_t load_idx = (is_encoder_cu) ? enc_load_idx : dev_idx;
        bool    is_xav1  = false;
        // job_cu_info[plugin_idx].kernel_name.size() is 3 for encoder CU and 1 for rest of CUs
        for (size_t i = 0; i < job_cu_info[plugin_idx].kernel_name.size(); ++i) {
          // enc_load is a 3 value pair (is_xav1 flag, Encoder Load, LA load)
          if (is_encoder_cu && 0 == i) {
            is_xav1 = job_cu_info[plugin_idx].cu_load[load_idx + i];
            continue;
          }
          // populate cu info details for a non zero load
          if (job_cu_info[plugin_idx].cu_load[load_idx + i] > 0) {
            fprintf(stdout, "\n%s [%u]: %u", job_cu_info[plugin_idx].kernel_name[i].c_str(), channel_count, job_cu_info[plugin_idx].cu_load[load_idx + i]);
            fill_xrm_cu_props(app_info_ptr->job_info[job_index].xrm_cu_pool_prop.get(), job_cu_info[plugin_idx].cu_load[load_idx + i], index,
                job_cu_info[plugin_idx].kernel_name[i - is_xav1], job_cu_info[plugin_idx].kernel_alias[i - is_xav1], dev_idx);
            // XRM 2nd drop integration - Logical Group Indexing support
            // Reset LG contraint info
            app_info_ptr->job_info[job_index].xrm_cu_pool_prop->cuListProp.cuProps[index].logicGroupInfo.constraintInfo = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_NULL;
            // update LG info for Encoder CUs only
            if (is_encoder_cu) {
              update_lg_info(app_info_ptr->job_info[job_index].xrm_cu_pool_prop.get(), job_cu_info[plugin_idx].cu_load[load_idx + i],
                  app_info_ptr->job_info[job_index].is_ull, index, job_cu_info[plugin_idx].kernel_alias[i - is_xav1], job_cu_info[plugin_idx].cu_load[load_idx + 2],
                  pseudo_logic_group_idx);
            }
            is_xav1 = false;
            ++index;
          }
        }
        --pseudo_logic_group_idx;

        if (is_encoder_cu)
          enc_load_idx += job_cu_info[plugin_idx].kernel_name.size();
      }
    }
  }
  app_info_ptr->job_info[job_index].xrm_cu_pool_prop->cuListProp.cuNum = index;
  return;
}

// Performs load calculation for a given job type
static int calc_job_load(jobSlotAppInfo* app_info_ptr, int32_t job_index, pt::ptree job) {
  xrmPluginFuncParam* param = app_info_ptr->plugin_param_ptr.get();
  memset(param, 0, sizeof(xrmPluginFuncParam));
  std::stringstream job_str;
  boost::property_tree::write_json(job_str, job);

  // Copy json string(job info) into param->input to send it to XRM plugins
  strncpy(param->input, job_str.str().c_str(), job_str.str().length());

  // Check for each plugin at a time
  for (uint32_t plugin_idx = 0; plugin_idx < PLUGIN_COUNT; ++plugin_idx) {
    cuInfo* job_cu_info = app_info_ptr->cu_info;

    // Reset cu_load and num_channels for next job
    job_cu_info[plugin_idx].cu_load.clear();
    job_cu_info[plugin_idx].num_channels.clear();

    // Invoke the respective XRM plugin for load calculation
    if (xrmExecPluginFunc(app_info_ptr->xrm_ctx, const_cast<char*>(job_cu_info[plugin_idx].plugin_name.c_str()), XRM_PLUGIN_FUNCID, param) != XRM_SUCCESS) {
      fprintf(stderr, "\nError: %s plugin function= %u fail to run the function\n", job_cu_info[plugin_idx].kernel_name[0].c_str(), XRM_PLUGIN_FUNCID);
      return 1;
    } else {
      char* endptr;
      char* token;
#ifdef DEBUG
      fprintf(stdout, "\nPlugin : %s\n", job_cu_info[plugin_idx].kernel_name[0].c_str());
      fprintf(stdout, "\nOutput param is : %s\n", param->output);
#endif
      errno = 0;
      token = static_cast<char*>(strtok(param->output, " "));
      if (token == NULL)
        break;

      app_info_ptr->job_info[job_index].num_dev_per_cmd = static_cast<uint32_t>(strtol(token, &endptr, 0));
      if (errno != 0) {
        fprintf(stderr, "\nError: String to long conversion error / Load token not found\n");
        return 1;
      } else if (app_info_ptr->job_info[job_index].num_dev_per_cmd > 1) {
        fprintf(stderr, "\nError: Multi-device(%u) usecases are not supported\n", app_info_ptr->job_info[job_index].num_dev_per_cmd);
        return 1;
      }

      // loop count indicates device count
      for (uint32_t nd = 0; nd < app_info_ptr->job_info[job_index].num_dev_per_cmd; ++nd) {
        uint32_t num_channels = static_cast<int32_t>(strtol(static_cast<char*>(strtok(NULL, " ")), &endptr, 0));
        if (errno != 0) {
          fprintf(stderr, "\nError: String to long conversion error / no of channels token not found \n");
          return 1;
        }

        job_cu_info[plugin_idx].num_channels.push_back(num_channels);

        // Loop count indicates no of channels for each plugin type
        for (uint32_t i = 0; i < num_channels; ++i) {
          // Channel loop > 1 valid for Encoder CUs only
          for (size_t j = 0; j < job_cu_info[plugin_idx].kernel_name.size(); ++j) {
            // get the CU loads
            uint32_t cu_load = static_cast<int32_t>(strtol(static_cast<char*>(strtok(NULL, " ")), &endptr, 0));
            cu_load          = cu_load * app_info_ptr->job_info[job_index].load_factor;
            // check for strtol errors
            if (errno != 0) {
              fprintf(stderr, "\nError: String to long conversion error / Load token not found\n");
              return 1;
            }
            if (cu_load > XRM_MAX_CU_LOAD_GRANULARITY_1000000 && !(app_info_ptr->job_info[i].is_ull && job_cu_info[plugin_idx].kernel_name[0] == "encoder")) {
              fprintf(stderr, "\nError: Requested %s load = %u exceeds maximum capacity.\n", job_cu_info[plugin_idx].kernel_name[0].c_str(), cu_load);
              return 1;
            }
            job_cu_info[plugin_idx].cu_load.push_back(cu_load);
          }
        }
      }
    }
  }
  // Fill xrm cu pool property structure as per the CU loads
  fill_xrm_props(app_info_ptr, job_index);
  return 0;
}

// Performs load calculation for all job types
static int calc_all_job_load(jobSlotAppInfo* app_info_ptr) {
  // loop over job types - heterogenous support
  for (size_t i = 0; i < app_info_ptr->job_info.size(); ++i) {
    // Parse num_job_slots value if present in json file
    pt::ptree job;
    try {
      pt::read_json(app_info_ptr->job_info[i].json_file_name.c_str(), job);
      auto pr_ptree = job.get_child("request.parameters");
      // Overwrite num_job_slots variable, if num_job_slots specified explicitly in json
      if (pr_ptree.not_found() != pr_ptree.find("num_job_slots"))
        app_info_ptr->job_info[i].req_job_slots = pr_ptree.get<int32_t>("num_job_slots");
      // Overwrite if specified load_factor explicitly - to limit maximum allowed jobslots per device
      if (pr_ptree.not_found() != pr_ptree.find("load_factor")) {
        float load_factor = pr_ptree.get<float>("load_factor");
        // not allowing load_factor < 1.0 which will lead to over subscription
        if (load_factor > 1.0f) {
          app_info_ptr->job_info[i].load_factor = load_factor;
        } else if (load_factor < 1.0f) {
          fprintf(stdout, "Found load_factor(%0.2f) < 1.0f, defaults to 1.0f\n", load_factor);
        }
      }
      // For GA1.0, dual-pass, single slice ULL only supported and the resource requirements this mode is same Normal latency.
      // So no special handling is needed for resource management
      // if (pr_ptree.not_found() != pr_ptree.find("latency_mode") && (strcmp("ull", pr_ptree.get<std::string>("latency_mode").c_str()) == 0))
      //   app_info_ptr->job_info[i].is_ull = true;
    } catch (std::exception& e) {
      fprintf(stderr, "\nError: Exception %s While reading json File\n", e.what());
      return 1;
    }
    fprintf(stdout, "\nJob Description File: %s\n", app_info_ptr->job_info[i].json_file_name.c_str());
    if (app_info_ptr->job_info[i].load_factor > 1.0f) {
      fprintf(stdout, "Load Factor: %0.2f applied on calculated CU loads to limit maximum possible jobslots per device\n", app_info_ptr->job_info[i].load_factor);
    }
    fprintf(stdout, "Requested Loads:");
    // Calculate load for each job type
    if (calc_job_load(app_info_ptr, i, job)) {
      fprintf(stderr, "\nError: Job load calculation error in %s", app_info_ptr->job_info[i].json_file_name.c_str());
      return 1;
    }
  }
#ifdef DEBUG
  fprintf(stdout,
      "\n###############################################################"
      "\n\n");
#endif
  return 0;
}

static int initialize_jobslot_app(jobSlotAppInfo* app_info_ptr) {
  // Create XRM context
  app_info_ptr->xrm_ctx = static_cast<xrmContext*>(xrmCreateContext(XRM_API_VERSION_1));
  if (nullptr == app_info_ptr->xrm_ctx) {
    fprintf(stderr, "\nError: Xrm context creation failed\n");
    return 1;
  }

  // Allocate memory for plugin input & output params
  try {
    app_info_ptr->plugin_param_ptr = std::make_unique<xrmPluginFuncParam>();
  } catch (...) {
    fprintf(stderr, "\nError: memory allocation failed for plugin_param_ptr");
    return 1;
  }

  for (size_t i = 0; i < app_info_ptr->job_info.size(); ++i) {
    app_info_ptr->job_info[i].req_job_slots = std::min(static_cast<size_t>(app_info_ptr->job_info[i].req_job_slots), app_info_ptr->job_info[i].input_files.size());
    try {
      app_info_ptr->job_info[i].xrm_cu_pool_prop = std::make_unique<xrmCuPoolPropertyV2>();
    } catch (...) {
      fprintf(stderr, "\nError: memory allocation failed for xrm_cu_pool_prop\n");
      return 1;
    }
    memset(app_info_ptr->job_info[i].xrm_cu_pool_prop.get(), 0, sizeof(xrmCuPoolPropertyV2));
  }
  return 0;
}

// Update reservation Id to the sh file created
static int update_xrm_reservation_ids(jobSlotAppInfo* app_info_ptr) {
  // fprintf (stdout,  "\nFunction: %s Start", __func__ );
  int32_t           total_pending_job_types = 0;
  std::vector<bool> no_resource_for_job_type;
  for (size_t i = 0; i < app_info_ptr->job_info.size(); ++i) {
    no_resource_for_job_type.push_back(false);
    if (app_info_ptr->job_info[i].allot_job_slots != app_info_ptr->job_info[i].req_job_slots) {
      ++total_pending_job_types;
    }
  }

  if (total_pending_job_types == 0)
    return 0;

  std::string                          xrm_ch_id, xrm_dev_id;
  xrmCuResInforV2*                     cu_res;
  xrmCuListResInforV2*                 cu_list_res;
  std::unique_ptr<xrmCuPoolResInforV2> xrm_cu_pool_res;

  try {
    xrm_cu_pool_res = std::make_unique<xrmCuPoolResInforV2>();
  } catch (...) {
    fprintf(stderr, "\nError: memory allocation failed for app_info_ptrl\n");
    return 1;
  }
  memset(xrm_cu_pool_res.get(), 0, sizeof(xrmCuPoolResInforV2));
  uint32_t last_res_id_loc = 0;
  while (true) {
    // Traverse for all job type
    for (size_t lc = app_info_ptr->last_res_id; lc < app_info_ptr->job_info.size() + app_info_ptr->last_res_id; ++lc) {
      uint32_t i = lc % static_cast<uint32_t>(app_info_ptr->job_info.size());
      xrm_ch_id.clear();
      if ((app_info_ptr->job_info[i].allot_job_slots == app_info_ptr->job_info[i].req_job_slots || true == no_resource_for_job_type[i])) {
        continue;
      }
      // printf("job type %d\n", (int)i);
      // Get Reservation id
      int32_t reservation_id = xrmCuPoolReserveV2(app_info_ptr->xrm_ctx, app_info_ptr->job_info[i].xrm_cu_pool_prop.get(), xrm_cu_pool_res.get());
      if (reservation_id > 0) {
        if (app_info_ptr->log_level > 1) {
          fprintf(stdout, "\nJob type (%d) got res id %d\n", (int) i, reservation_id);
        }
        // Push resrvation id to the vector and mark it as unused
        app_info_ptr->job_info[i].reservation_ids.push_back(reservation_id);
        app_info_ptr->job_info[i].reservation_id_status.push_back(RESERVE_ID_STATUS_UNUSED);
        try {
          app_info_ptr->job_info[i].thread_data_ptr.push_back(std::make_unique<thread_data>());
          app_info_ptr->job_info[i].thread_ptr.push_back(std::make_unique<pthread_t>());
        } catch (...) {
          fprintf(stderr, "\nError: memory allocation failed for thread_data\n");
          return 1;
        }
        // Modification done //
        xrm_dev_id.clear();
        xrm_dev_list pop_dev_list;
        for (uint32_t nd = 0; nd < app_info_ptr->job_info[i].num_dev_per_cmd; ++nd) {
          cu_list_res = &(xrm_cu_pool_res->cuListResInfor[0]);
          cu_res      = &(cu_list_res->cuResInfor[app_info_ptr->job_info[i].start_cu_idx[nd]]);

          // fprintf (stdout, "\nDevice id_reserved is : %d\n", cu_res->deviceId);
          pop_dev_list.hw_dev_ids.push_back(cu_res->deviceId);
          pop_dev_list.dev_start_cuidx.push_back(app_info_ptr->job_info[i].start_cu_idx[nd]);
        }
        if (!get_ffmpeg_cmd_vt_devs(app_info_ptr->job_info[i].run_commmand, &pop_dev_list)) {
          app_info_ptr->job_info[i].pop_dev_list_vec.push_back(pop_dev_list);
        } else {
          continue;
        }
        // untill here //
        ++app_info_ptr->job_info[i].allot_job_slots;
        ++app_info_ptr->total_jobs_reserved;
        last_res_id_loc = i;
      } else {
        no_resource_for_job_type[i] = true;
      }
      // Decrement total job count if requested jobs is equal to alloted or reservation id is negative
      if (app_info_ptr->job_info[i].allot_job_slots == app_info_ptr->job_info[i].req_job_slots || true == no_resource_for_job_type[i]) {
        --total_pending_job_types;
      }
    }
    if (total_pending_job_types <= 0)
      break;
  }
  app_info_ptr->last_res_id = (last_res_id_loc + 1) % static_cast<uint32_t>(app_info_ptr->job_info.size());
  return 0;
}

// separate a command line string into an array of arguments
int separate_cmdline(char* cmdline, char* cmd_params[CMD_SIZE]) {
  int index           = 0; // index of a command line we are reading from
  int param           = 0; // current parameter
  int param_index     = 0; // current character in the parameter
  int in_single_quote = 0; // are we in a single quote
  int in_double_quote = 0; // are we in a double quote
  while (cmdline[index] != '\0') {
    // skip spaces
    while ((cmdline[index] != '\0') && (isspace(cmdline[index])))
      index++;

    // Loop copying the parameter
    cmd_params[param] = &cmdline[param_index];
    while ((cmdline[index] != '\0') && ((!isspace(cmdline[index])) || in_single_quote || in_double_quote)) {
      if (cmdline[index] == '\'')
        in_single_quote = !in_single_quote;
      if (cmdline[index] == '\"')
        in_double_quote = !in_double_quote;
      if ((cmdline[index] != '\'') && (cmdline[index] != '\"')) {
        cmdline[param_index] = cmdline[index];
        ++param_index;
      }
      ++index;
    }
    if (in_single_quote || in_double_quote) {
      // error if we reached the end of the command line and are still in a quote
      fprintf(stderr, "\nCommand line error!\n");
      return -1;
    }
    cmdline[param_index] = '\0';
    ++param_index;
    ++param;
    ++index;
  }
  cmd_params[param] = NULL;
  return 0;
}

// Insert file name with numbers for each output file in cmd
static int insert_out_file_with_ch(uint32_t ch_str, std::string& token) {
  if (token.find_last_of(".") != std::string::npos) {
    uint64_t ns_time = (std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::system_clock::now().time_since_epoch()).count()) % (10000 * 10000);
    token.insert(token.find_last_of("."), "_" + std::to_string(ch_str) + "_" + std::to_string(ns_time));
  }
  return 0;
}

// Get how many numbers are there just after found index next characters
int get_no_of_digits_next(const std::string& device_string2, int32_t found) {
  int no_of_digit = 0;
  for (size_t i = found + 10; i < device_string2.size(); i++) {
    if (device_string2[i] >= '0' && device_string2[i] <= '9') {
      no_of_digit++;
    } else {
      break;
    }
  }
  return no_of_digit;
}

// Get ffmpeg command with modified reservation id and input file
int prepare_ffmpeg_run_cmd(const std::string& source_filename, int channel_id, std::string& run_cmd, xrm_dev_list* pop_dev_list) {
  // fprintf (stdout, "\nInput Run command :%s\n", run_cmd.c_str());
  std::string        token, input_source_filename, cmd;
  std::istringstream tokenStream(run_cmd);
  std::istringstream inputStream(source_filename);
  while (std::getline(tokenStream, token, ' ')) {
    (void) pop_dev_list;
    if (token == "-i") {
      if (std::getline(inputStream, input_source_filename, ' ')) {
        token += " " + input_source_filename;
      } else {
        fprintf(stderr, "\nError: Not enough input source files provided\n");
        return 1;
      }
    } else if (token == "-y") {
      cmd = cmd + token + " ";
      std::getline(tokenStream, token, ' ');
      insert_out_file_with_ch(channel_id, token);
    }
    cmd = cmd + token + " ";
  }
  run_cmd = cmd.substr(0, cmd.find_last_not_of(" \n\r\t\f\v") + 1);
  // fprintf (stdout, "\nModified Run command :%s\n", run_cmd.c_str());
  return 0;
}

// Get gstreamer command with modified reservation id and input file
int prepare_gst_run_cmd(const std::string& source_filename, int channel_id, std::string& run_cmd, xrm_dev_list* pop_dev_list) {
  // fprintf (stdout, "\nInput Run command :%s\n", run_cmd.c_str());
  uint32_t           sink_count = 0;
  std::string        token, input_source_filename, cmd;
  std::istringstream tokenStream(run_cmd);
  std::istringstream inputStream(source_filename);
  while (std::getline(tokenStream, token, ' ')) {
    (void) pop_dev_list;
    if (token == "filesrc") {
      if (std::getline(inputStream, input_source_filename, ' ')) {
        token += " location=" + input_source_filename + " ";
        cmd = cmd + token + " ";
      } else {
        fprintf(stderr, "\nError: Not enough input source files provided\n");
        return 1;
      }
    } else if (token == "filesink") {
      cmd = cmd + token + " ";
      std::getline(tokenStream, token, ' ');
      insert_out_file_with_ch(channel_id, token);
      cmd = cmd + token + " ";
    } else if (token == "fpsdisplaysink") {
      // fpsdisplay_token1 needs to be inserted in the modififed run command, if there is no "name=****" in the input command for fpsdisplaysink name
      std::string fpsdisplay_token1 = " name=fpsdisplaysink_" + std::to_string(channel_id) + "_" + std::to_string(sink_count++) + " ";
      std::string fpsdisplay_token2;
      cmd = cmd + token + " ";
      while (std::getline(tokenStream, token, ' ')) {
        if (token.starts_with("name=")) {
          token += "_" + std::to_string(channel_id);
          fpsdisplay_token1 = "";
        }
        fpsdisplay_token2 += token + " ";
        if (token == "!") {
          break;
        }
      }
      cmd += fpsdisplay_token1 + fpsdisplay_token2;
    } else
      cmd = cmd + token + " ";
  }
  run_cmd = cmd.substr(0, cmd.find_last_not_of(" \n\r\t\f\v") + 1);
  // fprintf (stdout, "\nModified Run command :%s\n", run_cmd.c_str());
  return 0;
}

// Get XMA apps command with modified reservation id and input file
int prepare_xma_apps_run_cmd(const std::string& source_filename, int channel_id, std::string& run_cmd, xrm_dev_list* pop_dev_list) {
  // fprintf (stdout, "\nInput Run command :%s\n", run_cmd.c_str());
  std::string        token, input_source_filename, cmd;
  std::istringstream tokenStream(run_cmd);
  std::istringstream inputStream(source_filename);
  while (std::getline(tokenStream, token, ' ')) {
    (void) pop_dev_list;
    if (token == "-i") {
      if (std::getline(inputStream, input_source_filename, ' ')) {
        token += " " + input_source_filename;
      } else {
        fprintf(stderr, "\nError: Not enough input source files provided\n");
        return 1;
      }
    } else if (token == "-o") {
      cmd = cmd + token + " ";
      std::getline(tokenStream, token, ' ');
      insert_out_file_with_ch(channel_id, token);
    }
    cmd = cmd + token + " ";
  }
  run_cmd = cmd.substr(0, cmd.find_last_not_of(" \n\r\t\f\v") + 1);
  // fprintf (stdout, "\nModified Run command :%s\n", run_cmd.c_str());
  return 0;
}

int thread_callback(void* opaque, int job_type, int reservation_id) {
  // fprintf (stdout,"\n%s jt = %d, rid = %d\n", "thread_callback", job_type, reservation_id);
  jobSlotAppInfo* app = static_cast<jobSlotAppInfo*>(opaque);
  pthread_mutex_lock(&app->job_info[job_type].mutex_lock);
  auto itr = find(app->job_info[job_type].reservation_ids.begin(), app->job_info[job_type].reservation_ids.end(), reservation_id);
  if (itr == app->job_info[job_type].reservation_ids.end()) {
    fprintf(stderr, "\nError: Unable to find resev_idx(%d) in reservation_ids list for job_type(%d)", reservation_id, job_type);
    pthread_mutex_unlock(&app->job_info[job_type].mutex_lock);
    return -1;
  }
  int idx = std::distance(app->job_info[job_type].reservation_ids.begin(), itr);
  // printf("%s idx = %d\n", "thread_callback", idx);
  app->job_info[job_type].reservation_id_status[idx] = RESERVE_ID_STATUS_FREE;
  pthread_mutex_unlock(&app->job_info[job_type].mutex_lock);
  return 0;
}

// Launch a process, wait for it to end, then release the allocated resources
void* launch_job(void(*launchargs)) {
  struct thread_data* thread_info    = static_cast<struct thread_data*>(launchargs);
  xrm_dev_list*       pop_dev_list   = thread_info->pop_dev_list;
  int                 log_level      = thread_info->log_level;
  uint32_t            reservation_id = thread_info->reservation_id;
  uint32_t            job_type       = thread_info->job_type;

  // fork new process
  int pid = fork();
  if (pid < 0) {
    fprintf(stderr, "\nError: worker.cpp::launch() Fork failed\n");
    return (void*) 1;
  }

  if (pid == 0) {
    // set environmant variable with reservation id
    char alloc_reservation_id[16];
    sprintf(alloc_reservation_id, "%u", reservation_id);
    setenv("XRM_RESERVE_ID", alloc_reservation_id, 1);
    // fprintf (stdout,  "\nxrm_reservation_id =%d\n", reservation_id );

    // write console output to seperate log files with '-enable-logging' option
    int file_out = -1;
    if (log_level) {
      if (check_folder_access())
        exit(1);
      char       fname[256];
      time_t     now;
      struct tm* tm;
      time(&now);
      tm = localtime(&now);
      sprintf(fname, "%sjobtype%u_resid%u_pid%d_%04d%02d%02d_%02d%02d%02d.log", LOG_FILE_PATH, job_type, reservation_id, getpid(), tm->tm_year + 1900, tm->tm_mon + 1,
          tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
      file_out = open(fname, O_RDWR | O_CREAT, 0644);
    }

    std::string cmdline_test_2 = thread_info->pre_src_cmdline;
    if (thread_info->pre_src_cmdline.find("ffmpeg") != std::string::npos) { // Handling of FFMPEG run command
      // insert source file into command line and add Reservation ID to the output for each process
      if (prepare_ffmpeg_run_cmd(thread_info->source_filename, reservation_id, cmdline_test_2, pop_dev_list)) {
        fprintf(stderr, "\nError: Failed preapring ffmpeg comand \n");
        exit(EXIT_FAILURE);
      }
    } else if (thread_info->pre_src_cmdline.find("gst-launch-1.0") != std::string::npos) { // Handling of GStreamer run command
      if (prepare_gst_run_cmd(thread_info->source_filename, reservation_id, cmdline_test_2, pop_dev_list)) {
        fprintf(stderr, "\nError: Failed preapring gstreamer comand \n");
        exit(EXIT_FAILURE);
      }
    } else if (thread_info->pre_src_cmdline.find("ma35_") != std::string::npos) { // Handling of XMA apps run command
      if (prepare_xma_apps_run_cmd(thread_info->source_filename, reservation_id, cmdline_test_2, pop_dev_list)) {
        fprintf(stderr, "\nError: Failed preapring gstreamer comand \n");
        exit(EXIT_FAILURE);
      }
    } else {
      fprintf(stderr, "\nError : Unknown Executable \n\n");
      exit(EXIT_FAILURE);
    }

    // separate command line into array of arguments
    /* Need to dynamically calculate size of parameters */
    char cmdline_test[LAUNCHER_CMD_SIZE];
    memset(cmdline_test, 0, sizeof(cmdline_test));
    char* cmd_params[CMD_SIZE];
    memset(cmd_params, 0, sizeof(cmd_params));
    strcpy(cmdline_test, cmdline_test_2.c_str());
    if (separate_cmdline(cmdline_test, cmd_params))
      exit(1);

    char cmd[CMD_SIZE];
    strcpy(cmd, cmd_params[0]);
    const char* file = strrchr(cmd, '/');
    if (file)
      file++;
    else
      file = cmd;
    strcpy(cmd_params[0], file);
    if (log_level) {
      if (dup2(file_out, 0) < 0)
        fprintf(stderr, "\nUnable to copy log file descriptor to stdin\n");
      if (dup2(file_out, 1) < 0)
        fprintf(stderr, "\nUnable to copy log file descriptor to stdout\n");
      if (dup2(file_out, 2) < 0)
        fprintf(stderr, "\nUnable to copy log file descriptor to stderr\n");
      //dup2  0-> stdin  || 1-> stdout(like printf) || 2-> stderr
      close(file_out);
    }

    fprintf(stdout, "\n\nCommand Running : ");
    for (size_t i = 0; i < CMD_SIZE; i++) {
      if (cmd_params[i] != NULL)
        fprintf(stdout, "%s ", cmd_params[i]);
    }

    fprintf(stdout, "\n");
    // execute command
    if (execvp(cmd, cmd_params) != 0) {
      fprintf(stderr, "\nError: execvp() failed\n\t%s: %s\n", cmd, strerror(errno));
      exit(EXIT_FAILURE);
    }
    return (void*) 0;
  }
  // Added to handle CTRL+C while lancher application is active
  // Store PIds in a global vector for cleanup later
  // fprintf(stdout, "\nStore child process PID %d\n", pid);
  pthread_mutex_lock(&gl_mutex_lock);
  gl_pids.push_back(pid);
  pthread_mutex_unlock(&gl_mutex_lock);

  // wait for child to finish
  int wstatus;
  waitpid(pid, &wstatus, 0);

  // Clear PIds from global vector once child process completes processing
  // fprintf(stdout, "\nClear child process PID %d from global vector\n", pid);
  pthread_mutex_lock(&gl_mutex_lock);
  gl_pids.erase(std::remove(gl_pids.begin(), gl_pids.end(), pid), gl_pids.end());
  pthread_mutex_unlock(&gl_mutex_lock);

  if (log_level > 1) {
    if (wstatus != 0) {
      fprintf(stdout, "\nJob failed for job_type(%d) res_id(%u) filename %s\n", (int) job_type, reservation_id, thread_info->source_filename.c_str());
    } else {
      fprintf(stdout, "\nJob passed for job_type(%d) res_id(%u)\n", (int) job_type, reservation_id);
    }
  }

  thread_info->callback(thread_info->opaque, (int) job_type, (int) reservation_id);

  if (WIFEXITED(wstatus)) {
    int status_code = WEXITSTATUS(wstatus);
    // fprintf (stdout,"\n Exit code is 0x%08x\nstatus_code 0x%08x\n", wstatus, status_code);
    return (void*) ((intptr_t) (status_code));
  }
  return (void*) 0;
}

bool join_job_thread(jobSlotInfo* curr_jtype_ptr, size_t resv_idx) {
  void* status;
  int   ret = pthread_join(*curr_jtype_ptr->thread_ptr[resv_idx], &status);
  pthread_mutex_lock(&curr_jtype_ptr->mutex_lock);
  curr_jtype_ptr->reservation_id_status[resv_idx] = RESERVE_ID_STATUS_UNUSED;
  pthread_mutex_unlock(&curr_jtype_ptr->mutex_lock);
  if (ret || (intptr_t) status != 0) {
    //fprintf (stderr,  "\n\n**Error Executing command %s\n Input file: %s\n Reservation id used: %d\n", curr_jtype_ptr->run_commmand.c_str(), curr_jtype_ptr->thread_data_ptr[resv_idx]->source_filename.c_str(), curr_jtype_ptr->thread_data_ptr[resv_idx]->reservation_id );
    curr_jtype_ptr->failed_jobs.push_back(curr_jtype_ptr->thread_data_ptr[resv_idx]->source_filename);
    return false;
  }
  return true;
}

bool launch_job_thread(jobSlotAppInfo* app_info_ptr, size_t jbt_idx, size_t resv_idx, uint32_t file_idx) {
  jobSlotInfo* curr_jtype_ptr = &app_info_ptr->job_info[jbt_idx];
  // Store thread data
  thread_data* thread_data_ptr     = curr_jtype_ptr->thread_data_ptr[resv_idx].get();
  thread_data_ptr->source_filename = curr_jtype_ptr->input_files[file_idx];
  thread_data_ptr->pop_dev_list    = &curr_jtype_ptr->pop_dev_list_vec[resv_idx];
  thread_data_ptr->pre_src_cmdline = curr_jtype_ptr->run_commmand;
  thread_data_ptr->log_level       = app_info_ptr->log_level;
  thread_data_ptr->job_type        = jbt_idx; // job type
  thread_data_ptr->reservation_id  = uint32_t(curr_jtype_ptr->reservation_ids[resv_idx]);
  thread_data_ptr->opaque          = app_info_ptr;
  thread_data_ptr->callback        = thread_callback;

  pthread_mutex_lock(&curr_jtype_ptr->mutex_lock);
  curr_jtype_ptr->reservation_id_status[resv_idx] = RESERVE_ID_STATUS_INUSE;
  pthread_mutex_unlock(&curr_jtype_ptr->mutex_lock);

  if (app_info_ptr->log_level > 1) {
    fprintf(stdout, "\nLaunching job_type(%u) res_id(%u) filenum %u\n", static_cast<uint32_t>(jbt_idx), thread_data_ptr->reservation_id, file_idx);
  }

  pthread_t* th_ptr = curr_jtype_ptr->thread_ptr[resv_idx].get();
  int        rc     = pthread_create(th_ptr, NULL, launch_job, static_cast<void*>(thread_data_ptr));
  if (rc) {
    fprintf(stderr, "\nError: pthread_create() exited with %d\n", rc);
    return false;
  }
  return true;
}

void print_failed_jobs(jobSlotAppInfo* app_info_ptr) {
  fprintf(stderr, "\n\nFailed Sources:-\n");
  for (size_t jbt_idx = 0; jbt_idx < app_info_ptr->job_info.size(); ++jbt_idx) {
    jobSlotInfo* curr_jtype_ptr = &app_info_ptr->job_info[jbt_idx];
    if (curr_jtype_ptr->failed_jobs.empty()) {
      continue;
    }
    fprintf(stderr, "Job_type[%d] CMD: %s\n", (int32_t) jbt_idx, curr_jtype_ptr->run_commmand.c_str());
    for (size_t idx = 0; idx < curr_jtype_ptr->failed_jobs.size(); ++idx) {
      fprintf(stderr, "\t%s\n", curr_jtype_ptr->failed_jobs[idx].c_str());
    }
  }
  return;
}

#ifdef DEBUG
void print_debug_info(int32_t total_launched_jobs, int32_t total_launch_failures, int32_t total_finished_jobs, int32_t total_failed_sources) {
  fprintf(stdout, "\n\nLaunched Threads:%4d ( Thread creation:- Failed:%4d,  Success:%4d ),  ", total_launched_jobs, total_launch_failures,
      total_launched_jobs - total_launch_failures);
  fprintf(stdout, "Joined Threads:%4d ( Status:- Failed:%4d,  Success:%4d ),  ", total_finished_jobs, total_failed_sources, total_finished_jobs - total_failed_sources);
  fprintf(stdout, "Active Threads:%4d\n", total_launched_jobs - total_launch_failures - total_finished_jobs);
  return;
}
#endif

void print_jobs_status(int32_t total_finished_jobs, int32_t total_failed_sources) {
  if (total_failed_sources) {
    fprintf(stdout, "\n\n%d source(s) completed\n", total_finished_jobs - total_failed_sources);
    fprintf(stdout, "%d source(s) failed!\n", total_failed_sources);
  } else {
    fprintf(stdout, "\n\nAll %d source(s) completed\n", total_finished_jobs - total_failed_sources);
  }
  return;
}

int launch_all_job(jobSlotAppInfo* app_info_ptr) {
  int32_t               total_input_sources   = 0;
  int32_t               total_launched_jobs   = 0;
  int32_t               total_failed_sources  = 0;
  int32_t               total_launch_failures = 0;
  int32_t               total_finished_jobs   = 0;
  std::vector<uint32_t> launched_jobs;

  fprintf(stdout, "\n\n");
  // check dev list for the given command and calulate no of input files for threads required.
  for (size_t i = 0; i < app_info_ptr->job_info.size(); ++i) {
    app_info_ptr->job_info[i].finished_jobs = 0;
    launched_jobs.push_back(0);
    total_input_sources += int32_t(app_info_ptr->job_info[i].input_files.size());
    fprintf(stdout, "JobType(%2d):- JobSlots: %2u, InputSources:%4d, JobDescriptionFile: \"%s\"\n", (int) i, app_info_ptr->job_info[i].allot_job_slots,
        (int) app_info_ptr->job_info[i].input_files.size(), app_info_ptr->job_info[i].json_file_name.c_str());
  }
  fprintf(stdout, "\nTotal Input Sources:%d\n", total_input_sources);
  fprintf(stdout, "\n===================================================================\n");
  if (app_info_ptr->log_level) {
    fprintf(stdout, "Launching processes started :\n");
    fprintf(stdout, "Log files per process are written to %s\n", LOG_FILE_PATH);
  }
  fprintf(stdout, "Launched max parallel process for the given job description. \nLeftover streams have to wait for free slots.\n");
  fprintf(stdout, "===================================================================\n\n");
  std::chrono::time_point<std::chrono::system_clock> start, end;
  start = std::chrono::system_clock::now();

  if (app_info_ptr->log_level) {
    if (check_folder_access()) {
      return EXIT_FAILURE;
    }
  }
  // Count failed sources and give error with failed command and input file
  while ((total_finished_jobs + total_launch_failures) < total_input_sources) {
    for (size_t jbt_idx = 0; jbt_idx < app_info_ptr->job_info.size(); ++jbt_idx) {
      jobSlotInfo* curr_jtype_ptr = &app_info_ptr->job_info[jbt_idx];
      if (curr_jtype_ptr->finished_jobs == curr_jtype_ptr->input_files.size() && curr_jtype_ptr->allot_job_slots == curr_jtype_ptr->req_job_slots)
        continue;
      // fprintf(stdout, "\njob_type " << jbt_idx << std::endl;
      for (size_t resv_idx = 0; resv_idx < curr_jtype_ptr->reservation_ids.size(); ++resv_idx) {
        pthread_mutex_lock(&curr_jtype_ptr->mutex_lock);
        reserveIdStatus reserve_id_status = curr_jtype_ptr->reservation_id_status[resv_idx];
        pthread_mutex_unlock(&curr_jtype_ptr->mutex_lock);
        // If reservation id is being used then continue
        if (reserve_id_status == RESERVE_ID_STATUS_INUSE) {
          continue;
        } else if (reserve_id_status == RESERVE_ID_STATUS_FREE) {
          if (false == join_job_thread(curr_jtype_ptr, resv_idx))
            total_failed_sources++;
          total_finished_jobs++;
          curr_jtype_ptr->finished_jobs++;
        }
      }
      // Check if all jobs are completed for a given job type, relinquish all channels and realloacte for other job types
      if (curr_jtype_ptr->finished_jobs == curr_jtype_ptr->input_files.size()) {
        curr_jtype_ptr->allot_job_slots = curr_jtype_ptr->req_job_slots;
        for (size_t resv_idx = 0; resv_idx < curr_jtype_ptr->reservation_ids.size(); ++resv_idx) {
          xrmCuPoolRelinquish(app_info_ptr->xrm_ctx, curr_jtype_ptr->reservation_ids[resv_idx]);
          --app_info_ptr->total_jobs_reserved;
        }
        curr_jtype_ptr->reservation_ids.clear();
        curr_jtype_ptr->reservation_id_status.clear();
        curr_jtype_ptr->pop_dev_list_vec.clear();
      }
      // fprintf (stdout,"curr_jtype_ptr[%d].finished_jobs %d\n", (int) jbt_idx, curr_jtype_ptr[jbt_idx].finished_jobs);
    }

    // Launch if reserve IDs are free and input source files are available
    if (total_launched_jobs < total_input_sources) {
      if (app_info_ptr->total_jobs_reserved == 0) {
        fprintf(stdout, "\nResource not available (Total Slots reserved - %d). Waiting for slots....\n", app_info_ptr->total_jobs_reserved);
      }
      for (size_t jbt_idx = 0; jbt_idx < app_info_ptr->job_info.size(); ++jbt_idx) {
        jobSlotInfo* curr_jtype_ptr = &app_info_ptr->job_info[jbt_idx];
        // continue with next job type if job launched for all input source files in the current job type
        if (launched_jobs[jbt_idx] == curr_jtype_ptr->input_files.size()) {
          continue;
        }
        if (curr_jtype_ptr->allot_job_slots < curr_jtype_ptr->req_job_slots) {
          update_xrm_reservation_ids(app_info_ptr);
        }
        for (size_t resv_idx = 0; resv_idx < curr_jtype_ptr->reservation_ids.size(); ++resv_idx) {
          // break if job launched for all input source files in the current job type
          if (launched_jobs[jbt_idx] == curr_jtype_ptr->input_files.size()) {
            break;
          }
          // Launch a new job if reservation id status is RESERVE_ID_STATUS_UNUSED
          pthread_mutex_lock(&curr_jtype_ptr->mutex_lock);
          reserveIdStatus reserve_id_status = curr_jtype_ptr->reservation_id_status[resv_idx];
          pthread_mutex_unlock(&curr_jtype_ptr->mutex_lock);
          if (reserve_id_status == RESERVE_ID_STATUS_UNUSED) {
            if (false == launch_job_thread(app_info_ptr, jbt_idx, resv_idx, launched_jobs[jbt_idx])) {
              // do we need to relaunch, if thread creation fails?
              // Right now, we capture failed cases and skip that file processing
              pthread_mutex_lock(&curr_jtype_ptr->mutex_lock);
              curr_jtype_ptr->reservation_id_status[resv_idx] = RESERVE_ID_STATUS_UNUSED;
              pthread_mutex_unlock(&curr_jtype_ptr->mutex_lock);
              curr_jtype_ptr->failed_jobs.push_back(curr_jtype_ptr->input_files[launched_jobs[jbt_idx]]);
              total_launch_failures++;
            }
            ++launched_jobs[jbt_idx];
            ++total_launched_jobs;
          }
        }
        // fprintf (stdout,"Launched Jobs[%d] %d\n", (int)jbt_idx, launched_jobs[jbt_idx]);
      }
    }
#ifdef DEBUG
    print_debug_info(total_launched_jobs, total_launch_failures, total_finished_jobs, total_failed_sources);
#endif
    if (total_failed_sources) {
      fprintf(stdout, "\r%d source(s) completed, %d source(s) failed!", total_finished_jobs - total_failed_sources, total_failed_sources);
    } else {
      fprintf(stdout, "\r%d source(s) completed", total_finished_jobs);
    }
    fflush(stdout);
    sleep(2);
  }
  end                                           = std::chrono::system_clock::now();
  std::chrono::duration<double> elapsed_seconds = end - start;
  long long                     min             = double(elapsed_seconds.count()) / 60.0;
  double                        sec             = double(elapsed_seconds.count() - min * 60.0);

  if (total_failed_sources) {
    print_failed_jobs(app_info_ptr);
  }

  print_jobs_status(total_finished_jobs, total_failed_sources);
  fprintf(stdout, "\nTime taken: %lld min %lf sec\n", min, sec);
  return 0;
}

// Clear all the child processes for CTRL+C signal
void cleanup_pids(void) {
  fprintf(stdout, "\nStarted cleanup of child processes\n");
  // Send SIGINT signal to all the active child processes
  pthread_mutex_lock(&gl_mutex_lock);
  for (size_t i = 0; i < gl_pids.size(); i++) {
    pid_t pid = gl_pids[i];
    // fprintf(stdout, "\npids[%ld] : %d\n", i, pid);
    // check if process is active
    if (kill(pid, 0) == 0) {
      // fprintf(stdout, "\nFound pid(%d) active\n", pid);
      kill(pid, SIGINT);
    }
#ifdef DEBUG
    else {
      fprintf(stdout, "\nFound pid(%d) already exited\n", pid);
    }
#endif
  }
  pthread_mutex_unlock(&gl_mutex_lock);

  // Wait for 2secs
  sleep(2);

  // Send SIGKILL signal to all the child processes which didnt respond to SIGINT signal
  pthread_mutex_lock(&gl_mutex_lock);
  for (size_t i = 0; i < gl_pids.size(); i++) {
    pid_t pid = gl_pids[i];
    // check if process is still active, i.e. unresponsive state
    if (kill(pid, 0) == 0) {
      // fprintf(stdout, "\t- Cleared unresponsive child process(pid-%d)\n", pid);
      kill(pid, SIGKILL);
    }
  }
  gl_pids.clear();
  // fprintf(stdout, "gl_pids vector cleared (size - %ld)\n", gl_pids.size());
  pthread_mutex_unlock(&gl_mutex_lock);
  return;
}

// Signal handler for CTRL+C
void my_sigint_handler(const int signal) {
  fprintf(stdout, "\nCaught CTRL+C signal %d\n", signal);
  cleanup_pids();
  exit(1);
}

int main(int argc, const char** argv) {

  //Get Terminal info of the current terminal
  struct termios terminfo;
  bool           got_term_attr;
  got_term_attr = true;
  if (tcgetattr(STDOUT_FILENO, &terminfo) != 0) {
    perror("tcgetattr() error");
    got_term_attr = false;
  }
  std::unique_ptr<jobSlotAppInfo> app_info_ptr;
  try {
    app_info_ptr = std::make_unique<jobSlotAppInfo>();
  } catch (...) {
    fprintf(stderr, "\nError: memory allocation failed for app_info_ptr\n");
    return EXIT_FAILURE;
  }

  // Parse & Validate command-line arguments
  if (parse_arguments(argc, argv, app_info_ptr.get())) {
    return EXIT_FAILURE;
  }

  // Initilaize data structures needed and create XRM context, reservation Id
  // bash files for each job type
  if (initialize_jobslot_app(app_info_ptr.get())) {
    return EXIT_FAILURE;
  }

  // Calculate load for each job type
  if (calc_all_job_load(app_info_ptr.get())) {
    return EXIT_FAILURE;
  }

  // update resevation id on sh files
  if (update_xrm_reservation_ids(app_info_ptr.get())) {
    return EXIT_FAILURE;
  }

  // sigint handler registered to handle CTRL+C
  signal(SIGINT, my_sigint_handler);

  // call launcher for each jobtype with every input provided.
  launch_all_job(app_info_ptr.get());

  //Reset the Terminal back to the intial state
  int sys_ret = 0;
  if (got_term_attr) {
    if (tcsetattr(STDOUT_FILENO, TCSADRAIN, &terminfo) != 0) {
      sys_ret = system("reset");
    }
  } else {
    sys_ret = system("reset");
  }
  if (sys_ret == -1) {
    fprintf(stderr, "\nError: While resetting the terminal\n");
  }

  return 0;
}
