// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#pragma once

#include <string>
#include <xrm.h>
#include "xrm_interface.h"

#define XRM_UTILS_PRECISION_1000000_BIT_MASK(load) ((load << 8))

typedef struct {
  std::string ip_name;
  std::string plugin_name;
  std::string kernel_name;
  std::string kernel_alias;
} XrmUtilsIPMapping;

/**
 * @brief Checks and allocates resources for xrm given the cu pool properties
 * with relevant device info
 * 
 * @param xrm_ctx The xrm context used for xrm calls
 * @param pool_property Pool property containing relevant device / job info
 * @param xrm_cu_list_res Will be filled by xrmCuListAllocV2
 * @return XRM_SUCCESS or XRM_ERROR
 */
int xrm_utils_reserve_resources(xrmContext xrm_ctx, xrmCuPoolPropertyV2* pool_property, xrmCuListResourceV2* xrm_cu_list_res);

/**
 * @brief Update the cu pool properties based on the relevant information
 * provided
 * 
 * @param ip_name What module this is: "DECODER", "SCALER" or "ENCODER"
 * @param device_id The device index whose resources we should query
 * @param cu_load The load of the job
 * @param cu_pool_prop The properties to be set
 * @return XRM_SUCCESS or XRM_ERROR
 */
int xrm_utils_update_pool_props(const std::string& ip_name, int dev_index, int cu_load, xrmCuPoolPropertyV2* cu_pool_prop, bool is_xav1);

/**
 * @brief Calculate the load for the module specified by ip_name. Can be
 * "DECODER", "SCALER", or "ENCODER"
 * 
 * @param xrm_ctx xrm context used for calls to xrm
 * @param xrm_props Values relevant for creating a json for MA35 load calculation
 * @param input_props Props used to calculate load
 * @param output_props Props used to calculate load (if the output affects the 
 * density such as scaler)
 * @param ip_name Can be "DECOER", "SCALER" or "ENCODER"
 * @param cu_load Pointer to the cu load to be set, in the case of encoder, this
 * is an array of size 3.
 * cu_load[0] = is_xav1 flag
 * cu_load[1] = encoder load
 * cu_load[2] = lookahead load
 * @return The cu_load calculated, or XRM_ERROR on error
 */
int xrm_utils_load_calc(xrmContext xrm_ctx, XrmInterfaceProperties* input_props, XrmInterfaceProperties* output_props, int num_outputs, const std::string& ip_name,
    int* cu_loads, bool is_xav1);
